ALL_CASES = [
    "ガ",
    "デ",
    "ト",
    "ニ",
    "ノ",
    "ヘ",
    "ヲ",
    "カラ",
    "ガ２",
    "ノ？",
    "マデ",
    "ヨリ",
    "トイウ",
    "トシテ",
    "トスル",
    "ニオク",
    "ニシテ",
    "ニツク",
    "ニトル",
    "ニヨル",
    "マデニ",
    "ニオイテ",
    "ニカワル",
    "ニソッテ",
    "ニツイテ",
    "ニトッテ",
    "ニムケテ",
    "ニムケル",
    "ニヨッテ",
    "ニヨラズ",
    "ニアワセテ",
    "ニカギッテ",
    "ニカギラズ",
    "ニカランデ",
    "ニカワッテ",
    "ニカンシテ",
    "ニカンスル",
    "ニクラベテ",
    "ニクワエテ",
    "ニタイシテ",
    "ニタイスル",
    "ニツヅイテ",
    "ニナランデ",
    "ヲツウジテ",
    "ヲツウジル",
    "ヲノゾイテ",
    "ヲフクメテ",
    "ヲメグッテ",
    "ニトモナッテ",
    "ニモトヅイテ",
    "無",
    "修飾",
    "判ガ",
    "時間",
    "外の関係",
]
ALL_CASES += [case + "≒" for case in ALL_CASES]

ALL_COREFS = ["=", "=構", "=役"]
ALL_COREFS += [case + "≒" for case in ALL_COREFS]
